from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters

from wbhuman_resources.models import EmployeeHumanResource, Position


class PositionFilterSet(wb_filters.FilterSet):
    class Meta:
        model = Position
        fields = {"name": ["exact", "icontains"], "parent": ["exact"], "level": ["exact"], "manager": ["exact"]}


class BaseEmployeeFilterSet(wb_filters.FilterSet):
    contract_type = wb_filters.ChoiceFilter(
        label=_("Contract"),
        choices=EmployeeHumanResource.ContractType.choices,
        initial=EmployeeHumanResource.ContractType.INTERNAL,
    )
    is_active = wb_filters.BooleanFilter(label=_("Is Active"), initial=True)
    position = wb_filters.ModelChoiceFilter(
        label=_("Position"),
        queryset=Position.objects.all(),
        endpoint=Position.get_representation_endpoint(),
        value_key=Position.get_representation_value_key(),
        label_key=Position.get_representation_label_key(),
        method="filter_position",
        filter_params={"height": 0},
    )

    def filter_position(self, queryset, name, value):
        if value:
            return queryset.filter(position__in=value.get_descendants(include_self=True)).distinct()
        return queryset


class EmployeeBalanceFilterSet(BaseEmployeeFilterSet):
    took_long_vacations = wb_filters.BooleanFilter(field_name="took_long_vacations", lookup_expr="exact")

    extra_days_per_period__gte = wb_filters.NumberFilter(
        field_name="extra_days_per_period", method="filter_str_number_gte", lookup_expr="gte"
    )

    def filter_str_number_lte(self, queryset, name, value):
        if value:
            numerical_field_name = name.replace("_repr", "")
            return queryset.filter(**{f"{numerical_field_name}__lte": value})
        return queryset

    def filter_str_number_gte(self, queryset, name, value):
        if value:
            numerical_field_name = name.replace("_repr", "")
            return queryset.filter(**{f"{numerical_field_name}__gte": value})
        return queryset

    class Meta:
        model = EmployeeHumanResource
        fields = {"profile": ["exact"], "position": ["exact"], "extra_days_frequency": ["exact"]}


class EmployeeFilterSet(BaseEmployeeFilterSet):
    top_position_repr = wb_filters.ModelChoiceFilter(
        label=_("Position"),
        queryset=Position.objects.all(),
        endpoint=Position.get_representation_endpoint(),
        value_key=Position.get_representation_value_key(),
        label_key=Position.get_representation_label_key(),
        filter_params={"height": 1},
        method="filter_top_position",
    )

    def filter_top_position(self, queryset, name, value):
        if value:
            return queryset.filter(
                Q(position__parent__isnull=False) & Q(position__parent__in=value.get_descendants(include_self=True))
            ).distinct()
        return queryset

    class Meta:
        model = EmployeeHumanResource
        fields = {"calendar": ["exact"], "contract_type": ["exact"], "enrollment_at": ["gte", "lte"]}



from wbcore.menus import Menu

from .absence import ABSENCEREQUESTTYPE_MENUITEM
from .calendars import DAYOFF_MENUITEM, DAYOFFCALENDAR_MENUITEM
from .employee import POSITION_MENUITEM

ADMINISTRATION_MENU = Menu(
    label="Administration",
    items=[
        DAYOFF_MENUITEM,
        ABSENCEREQUESTTYPE_MENUITEM,
        DAYOFFCALENDAR_MENUITEM,
        POSITION_MENUITEM,
    ],
)



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbhuman_resources.models import AbsenceRequest, AbsenceRequestType


def get_type_legend():
    """
    Dynamically create the request type list legend
    """
    if AbsenceRequestType.objects.exists():
        legend = []
        for type in AbsenceRequestType.objects.all():
            try:
                legend.append(dp.LegendItem(icon=WBIcon[type.icon].icon, label=type.title, value=type.id))
            except KeyError:
                legend.append(dp.LegendItem(icon=type.icon, label=type.title, value=type.id))

        return dp.Legend(key="type", items=legend)


class AbsenceRequestDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="type_icon", label=" ", width=Unit.PIXEL(50)),
                dp.Field(key="employee", label=_("Employee"), width=Unit.PIXEL(200)),
                dp.Field(key="period", label=_("Period"), width=Unit.PIXEL(250)),
                dp.Field(key="_total_hours_in_days", label=_("Total Days"), width=Unit.PIXEL(125)),
                dp.Field(key="_total_vacation_hours_in_days", label=_("Vacation Days"), width=Unit.PIXEL(125)),
                dp.Field(key="department", label=_("Department"), width=Unit.PIXEL(350)),
                dp.Field(key="created", label=_("Created"), width=Unit.PIXEL(150)),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", AbsenceRequest.Status.DRAFT.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", AbsenceRequest.Status.PENDING.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", AbsenceRequest.Status.APPROVED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", AbsenceRequest.Status.DENIED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_DARK.value},
                            condition=("==", AbsenceRequest.Status.CANCELLED.name),
                        ),
                    ],
                ),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=AbsenceRequest.Status.DRAFT.label,
                            value=AbsenceRequest.Status.DRAFT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=AbsenceRequest.Status.PENDING.label,
                            value=AbsenceRequest.Status.PENDING.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=AbsenceRequest.Status.APPROVED.label,
                            value=AbsenceRequest.Status.APPROVED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=AbsenceRequest.Status.DENIED.label,
                            value=AbsenceRequest.Status.DENIED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_DARK.value,
                            label=AbsenceRequest.Status.CANCELLED.label,
                            value=AbsenceRequest.Status.CANCELLED.value,
                        ),
                    ],
                ),
                get_type_legend(),
            ],
        )

    def get_instance_display(self) -> Display:
        grid_fields = [[repeat_field(3, "status")], ["period", "employee", "type"]]
        try:
            if self.view.get_object().status == AbsenceRequest.Status.DENIED.name:
                grid_fields.append([repeat_field(3, "reason")])
        except AssertionError:
            pass

        grid_fields.extend(
            [
                ["attachment", repeat_field(2, "crossborder_country")],
                [repeat_field(3, "notes")],
                [repeat_field(3, "hours_section")],
                [repeat_field(3, "periods_section")],
            ]
        )
        return create_simple_display(
            grid_fields,
            [
                create_simple_section(
                    "hours_section",
                    _("Hours"),
                    [
                        ["_total_hours", "_total_hours_in_days"],
                        ["_total_vacation_hours", "_total_vacation_hours_in_days"],
                    ],
                    collapsed=False,
                ),
                create_simple_section("periods_section", _("Periods"), [["periods"]], "periods", collapsed=True),
            ],
        )


class AbsenceRequestTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="icon", label=_("Icon")),
                dp.Field(key="color", label=_("Color")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [
                    "title",
                    "icon",
                    "color",
                ],
                [repeat_field(3, "extra_notify_groups")],
                [repeat_field(3, "settings_section")],
                [repeat_field(3, "allowed_countries_section")],
            ],
            [
                create_simple_section(
                    "settings_section",
                    _("Settings"),
                    [
                        [
                            "is_vacation",
                            "is_timeoff",
                            "is_extensible",
                        ],
                        ["days_in_advance", "auto_approve", "is_country_necessary"],
                    ],
                    collapsed=False,
                ),
                create_simple_section(
                    "allowed_countries_section",
                    _("Allowed Countries (Cross-Border Rule)"),
                    [["crossbordercountries"]],
                    "crossbordercountries",
                    collapsed=False,
                ),
            ],
        )


class AbsenceRequestCrossBorderCountryDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["geography"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(fields=(dp.Field(key="geography_repr", label=_("Country")),))


class AbsenceRequestEmployeeHumanResourceDisplayConfig(AbsenceRequestDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="period", label=_("Period")),
                dp.Field(key="_total_hours_in_days", label=_("Total Days")),
                dp.Field(key="_total_vacation_hours_in_days", label=_("Vacation Days")),
                dp.Field(key="created", label=_("Created")),
            ]
        )


class AbsenceTypeCountEmployeeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="year", label=_("Year")),
                dp.Field(key="absence_type", label=_("Type")),
                dp.Field(key="hours_count", label=_("Count (hours)")),
                dp.Field(key="days_count", label=_("Count (days)")),
            ]
        )


class AbsenceRequestPeriodsAbsenceRequestDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="date", label=_("Date")),
                dp.Field(key="default_period", label=_("Period")),
                dp.Field(key="_total_hours", label=_("Total hours")),
                dp.Field(key="balance", label=_("Balance")),
                dp.Field(key="consecutive_hours_count", label=_("Consecutive hours count")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["date", repeat_field(2, "default_period")], ["_total_hours", "balance", "consecutive_hours_count"]]
        )


class AbsenceTablePandasDisplayConfig(DisplayViewConfig):
    FORMATTING_CELL_RULE = [
        dp.FormattingRule(
            style={"backgroundColor": "#BA324F", "color": "#BA324F"},
            condition=("==", -1),  # Absent
        ),
        dp.FormattingRule(
            style={"backgroundColor": "#F0F7F4", "color": "#F0F7F4"},
            condition=("==", 0),  # Present
        ),
        dp.FormattingRule(
            style={"backgroundColor": "#6A8D73", "color": "#6A8D73"},
            condition=("==", 1),  # Partially Present
        ),
        dp.FormattingRule(
            style={"backgroundColor": "#1E91D6", "color": "#1E91D6"},
            condition=("==", 2),  # Partially Remote
        ),
        dp.FormattingRule(
            style={"backgroundColor": "#18206F", "color": "#18206F"},
            condition=("==", 3),  # Remote
        ),
    ]

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="employee_repr",
                    label=_("Employee"),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"fontWeight": "bold"},
                            condition=("!=", None),
                        )
                    ],
                    width=Unit.PIXEL(200),
                ),
                dp.Field(key="position", label=_("Department"), width=Unit.PIXEL(300)),
                dp.Field(
                    key="monday", label=_("Monday"), formatting_rules=self.FORMATTING_CELL_RULE, width=Unit.PIXEL(150)
                ),
                dp.Field(
                    key="tuesday",
                    label=_("Tuesday"),
                    formatting_rules=self.FORMATTING_CELL_RULE,
                    width=Unit.PIXEL(150),
                ),
                dp.Field(
                    key="wednesday",
                    label=_("Wednesday"),
                    formatting_rules=self.FORMATTING_CELL_RULE,
                    width=Unit.PIXEL(150),
                ),
                dp.Field(
                    key="thursday",
                    label=_("Thursday"),
                    formatting_rules=self.FORMATTING_CELL_RULE,
                    width=Unit.PIXEL(150),
                ),
                dp.Field(
                    key="friday", label=_("Friday"), formatting_rules=self.FORMATTING_CELL_RULE, width=Unit.PIXEL(150)
                ),
                dp.Field(
                    key="saturday",
                    label=_("Saturday"),
                    formatting_rules=self.FORMATTING_CELL_RULE,
                    width=Unit.PIXEL(150),
                ),
                dp.Field(
                    key="sunday", label=_("Sunday"), formatting_rules=self.FORMATTING_CELL_RULE, width=Unit.PIXEL(150)
                ),
            ],
            legends=[
                dp.Legend(
                    items=[
                        dp.LegendItem(icon="#F0F7F4", label=_("Present")),
                        dp.LegendItem(icon="#6A8D73", label=_("Partially Present")),
                        dp.LegendItem(icon="#18206F", label=_("Remote")),
                        dp.LegendItem(icon="#1E91D6", label=_("Partially Remote")),
                        dp.LegendItem(icon="#BA324F", label=_("Absent")),
                    ]
                )
            ],
            # formatting=[
            #     dp.Formatting(
            #         column="employee_repr",
            #         formatting_rules=[
            #             dp.FormattingRule(
            #                 condition=("!=", None), style={"fontWeight": "bold"}
            #             )
            #         ]
            #     )
            # ]
        )



# Generated by Django 5.2.9 on 2025-12-09 08:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbhuman_resources', '0022_remove_review_editable_mode'),
    ]

    operations = [
        migrations.AlterField(
            model_name='employeehumanresource',
            name='weekly_off_periods',
            field=models.ManyToManyField(related_name='employees_off', through='wbhuman_resources.EmployeeWeeklyOffPeriods', through_fields=('employee', 'period'), to='wbhuman_resources.defaultdailyperiod', verbose_name='Weekly off periods'),
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-14 15:00

from django.db import migrations, models


def migrate_group(apps, schema_editor):
    Position = apps.get_model("wbhuman_resources", "Position")
    for position in Position.objects.filter(group__isnull=False):
        if position.group not in position.groups.all():
            position.groups.add(position.group)


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("wbhuman_resources", "0017_absencerequest_crossborder_country_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="position",
            name="groups",
            field=models.ManyToManyField(blank=True, related_name="human_resources_positions", to="auth.group"),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(migrate_group),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.RemoveField(
            model_name="position",
            name="group",
        ),
    ]



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class DayOffDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="date", label=_("Date")),
                dp.Field(key="count_as_holiday", label=_("Count as Holiday")),
                dp.Field(key="calendar", label=_("Calendar")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [[repeat_field(2, "title")], [repeat_field(2, "calendar")], ["date", "count_as_holiday"]]
        )


class DayOffCalendarDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="resource", label=_("Resource")),
                dp.Field(key="timezone", label=_("Timezone")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["title", "resource", "timezone"],
                [repeat_field(3, "default_periods_section")],
                [repeat_field(3, "days_off_section")],
            ],
            [
                create_simple_section(
                    "default_periods_section",
                    _("Default Periods"),
                    [["default_periods"]],
                    "default_periods",
                    collapsed=True,
                ),
                create_simple_section("days_off_section", _("Day off"), [["days_off"]], "days_off", collapsed=False),
            ],
        )


class DayOffDayOffCalendarDisplayConfig(DayOffDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="date", label=_("Date")),
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="count_as_holiday", label=_("Count as Holiday")),
            ]
        )


class DefaultDailyPeriodDayOffCalendarDisplayConfig(DisplayViewConfig):
    def get_list_display(self):
        return dp.ListDisplay(
            fields=[
                dp.Field(key="timespan", label=_("Time Range")),
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="total_hours", label=_("Total Hours")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["title", "title"], ["timespan", "total_hours"]])



from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Person

from wbhuman_resources.models import (
    Review,
    ReviewAnswer,
    ReviewGroup,
    ReviewQuestion,
    ReviewQuestionCategory,
)


class ReviewGroupFilter(wb_filters.FilterSet):
    class Meta:
        model = ReviewGroup
        fields = {
            "name": ["exact", "icontains"],
        }


def get_filter_params(request, view):
    return {"related_to": request.user.profile}


def filter_default_related_to(field, request, view):
    return request.user.profile.id if request.user.profile else None


class ReviewTemplateFilter(wb_filters.FilterSet):
    related_to = wb_filters.ModelChoiceFilter(
        label=_("Related to"),
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
        method="filter_related_to",
        initial=filter_default_related_to,
    )

    def filter_related_to(self, queryset, name, value):
        if value:
            queryset = queryset.filter(Q(reviewee=value) | Q(reviewer=value) | Q(moderator=value))
        return queryset

    class Meta:
        model = Review
        fields = {
            "review": ["lte", "gte"],
            "status": ["exact"],
            "moderator": ["exact"],
            "review_group": ["exact"],
            "changed": ["lte", "gte"],
        }


class ReviewFilter(ReviewTemplateFilter):
    is_template = wb_filters.BooleanFilter(initial=False, label=_("Is Template"))

    class Meta:
        model = Review
        fields = {
            "from_date": ["lte", "gte"],
            "to_date": ["lte", "gte"],
            "review_deadline": ["lte", "gte"],
            "review": ["lte", "gte"],
            "status": ["exact"],
            "reviewee": ["exact"],
            "reviewer": ["exact"],
            "moderator": ["exact"],
            "review_group": ["exact"],
            "is_template": ["exact"],
            "year": ["exact"],
            "type": ["exact"],
            "changed": ["lte", "gte"],
        }


class ReviewQuestionCategoryFilter(wb_filters.FilterSet):
    class Meta:
        model = ReviewQuestionCategory
        fields = {
            "name": ["exact", "icontains"],
            "order": ["exact"],
            "weight": ["exact"],
        }


class ReviewQuestionFilter(wb_filters.FilterSet):
    class Meta:
        model = ReviewQuestion
        fields = {
            "review": ["exact"],
            "category": ["exact"],
            "mandatory": ["exact"],
            "answer_type": ["exact"],
            "order": ["exact"],
            "weight": ["exact"],
        }


class ReviewAnswerFilter(wb_filters.FilterSet):
    question_name = wb_filters.CharFilter(label=_("Question"), lookup_expr="icontains")

    class Meta:
        model = ReviewAnswer
        fields = {"answer_text": ["exact", "icontains"], "answered_by": ["exact"], "question": ["exact"]}
        hidden_fields = ["question"]


class ReviewProgressReviewFilter(wb_filters.FilterSet):
    class Meta:
        model = ReviewAnswer
        fields = {}


class RatingReviewAnswerReviewFilter(wb_filters.FilterSet):
    deviation = wb_filters.ChoiceFilter(
        label=_("Deviation"),
        choices=[("EQUAL", _("Equal")), ("LESS", _("Less")), ("GREAT", _("Greater"))],
        method="fake_filter",
        clearable=False,
        required=False,
    )
    category_question_name = wb_filters.CharFilter(label=_("Category"), lookup_expr="icontains")
    question_name = wb_filters.CharFilter(label=_("Question"), lookup_expr="icontains")
    answer_type = wb_filters.ChoiceFilter(
        label="Answer Type", choices=ReviewQuestion.ANSWERTYPE.choices, required=False, hidden=True
    )

    class Meta:
        model = ReviewAnswer
        fields = {}



import pytest
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models.users import User
from wbcore.permissions.shortcuts import is_internal_user

from wbhuman_resources.factories.employee import EmployeeHumanResourceFactory
from wbhuman_resources.models.employee import EmployeeHumanResource

fake = Faker()


@pytest.mark.django_db
class TestPermissionTasks:
    @pytest.fixture
    def request_user_external(self):
        request = APIRequestFactory()
        user = UserFactory()
        EmployeeHumanResourceFactory.create(
            profile=user.profile, contract_type=EmployeeHumanResource.ContractType.EXTERNAL
        )
        request.user = user
        return request

    @pytest.fixture
    def request_user_active_internal(self):
        request = APIRequestFactory()
        user = UserFactory()
        EmployeeHumanResourceFactory.create(
            profile=user.profile, contract_type=EmployeeHumanResource.ContractType.INTERNAL
        )
        request.user = user
        return request

    @pytest.fixture
    def request_user_inactive_internal(self):
        request = APIRequestFactory()
        user = UserFactory()
        EmployeeHumanResourceFactory.create(
            profile=user.profile, contract_type=EmployeeHumanResource.ContractType.INTERNAL, is_active=False
        )
        request.user = user
        return request

    def test_permission_active_internal(self, request_user_active_internal):
        assert is_internal_user(request_user_active_internal.user) is True

    def test_permission_inactive_internal(self, request_user_inactive_internal):
        assert is_internal_user(request_user_inactive_internal.user) is False

    def test_permission_external(self, request_user_external):
        assert is_internal_user(request_user_external.user) is False

    def test_permission_external_but_considered_internal(self, request_user_external):
        from wbcore.permissions.registry import user_registry

        user = request_user_external.user
        global_preferences_registry.manager()["wbhuman_resources__is_external_considered_as_internal"] = True
        user = User.objects.get(id=user.id)  # reload to reset cached property
        user_registry.reset_cache()
        assert is_internal_user(user) is True



from datetime import date, timedelta

import pandas as pd
from django.utils.timezone import localdate
from django.utils.translation import gettext_lazy as _
from psycopg.types.range import TimestamptzRange
from wbcore import filters as wb_filters
from wbcore.contrib.agenda.filters import CalendarItemPeriodBaseFilterSet

from wbhuman_resources.models import (
    AbsenceRequest,
    AbsenceRequestPeriods,
    AbsenceRequestType,
    Position,
)


def current_year_date_start(*args, **kwargs):
    d = localdate()
    return date(d.year, 1, 1)


def current_year_date_end(*args, **kwargs):
    d = localdate()
    return max((d + pd.tseries.offsets.YearEnd(1)).date(), d + timedelta(days=60))


class AbsenceRequestFilter(CalendarItemPeriodBaseFilterSet):
    conference_room = boolean_conference_room = None
    department = wb_filters.ModelChoiceFilter(
        label=_("Department"),
        queryset=Position.objects.all(),
        endpoint=Position.get_representation_endpoint(),
        value_key=Position.get_representation_value_key(),
        label_key=Position.get_representation_label_key(),
        method="filter_position",
    )
    is_active_employee = wb_filters.BooleanFilter(
        label=_("Is Employee Active"), method="boolean_is_active_employee", initial=True
    )
    _total_hours_in_days__gte = wb_filters.NumberFilter(
        lookup_expr="gte", label="Total hours", field_name="_total_hours_in_days"
    )
    _total_vacation_hours_in_days__gte = wb_filters.NumberFilter(
        lookup_expr="gte", label="Total hours", field_name="_total_vacation_hours_in_days"
    )
    _total_hours_in_days__lte = wb_filters.NumberFilter(
        lookup_expr="lte", label="Total hours", field_name="_total_hours_in_days"
    )
    _total_vacation_hours_in_days__lte = wb_filters.NumberFilter(
        lookup_expr="lte", label="Total hours", field_name="_total_vacation_hours_in_days"
    )

    def get_default_period(self):
        return TimestamptzRange(lower=current_year_date_start(), upper=current_year_date_end())

    def boolean_is_active_employee(self, queryset, name, value):
        if value:
            return queryset.filter(employee__is_active=True)
        #     return queryset.filter(employee__in=EmployeeHumanResource.active_internal_employees.all())
        return queryset

    def filter_position(self, queryset, name, value):
        if value:
            return queryset.filter(department__in=value.get_descendants(include_self=True)).distinct()
        return queryset

    class Meta:
        model = AbsenceRequest
        fields = {
            "employee": ["exact"],
            "status": ["exact"],
            "type": ["exact"],
            "created": ["gte", "exact", "lte"],
        }


class AbsenceTypeCountEmployeeModelFilterSet(wb_filters.FilterSet):
    year = wb_filters.YearFilter(field_name="year", lookup_expr="exact")

    absence_type = wb_filters.ModelChoiceFilter(
        label=_("Type"),
        queryset=AbsenceRequestType.objects.all(),
        endpoint=AbsenceRequestType.get_representation_endpoint(),
        value_key=AbsenceRequestType.get_representation_value_key(),
        label_key=AbsenceRequestType.get_representation_label_key(),
        method="filter_absence_type",
    )

    def filter_absence_type(self, queryset, label, value):
        if value:
            return queryset.filter(request__type=value)
        return queryset

    class Meta:
        model = AbsenceRequestPeriods
        fields = {}


class AbsenceRequestEmployeeHumanResourceFilterSet(AbsenceRequestFilter):
    department = is_active_employee = None

    class Meta:
        model = AbsenceRequest
        fields = {
            "status": ["exact"],
            "type": ["exact"],
            "created": ["gte", "exact", "lte"],
        }



# Generated by Django 5.0.3 on 2024-04-11 08:24

import colorfield.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbhuman_resources", "0020_alter_employeeyearbalance_year_alter_review_year"),
    ]

    operations = [
        migrations.AlterField(
            model_name="position",
            name="color",
            field=colorfield.fields.ColorField(default="#FF0000", image_field=None, max_length=25, samples=None),
        ),
    ]



from django.db import connection
from django.dispatch import receiver
from wbcore.test.signals import (
    custom_update_kwargs,
    get_custom_factory,
    get_custom_serializer,
)

from wbhuman_resources.factories import (
    DefaultPersonKPIFactory,
    ReviewQuestionNoCategoryFactory,
    ReviewTemplateFactory,
)
from wbhuman_resources.models import AbsenceRequest, ReviewQuestion
from wbhuman_resources.serializers import KPIModelSerializer
from wbhuman_resources.viewsets import (
    AbsenceTablePandasViewSet,
    AbsenceTypeCountEmployeeModelViewSet,
    KPIModelViewSet,
    ReviewAnswerReviewNoCategoryModelViewSet,
    ReviewAnswerReviewPandasViewSet,
    ReviewAnswerReviewQuestionCategoryModelViewSet,
    ReviewModelViewSet,
    ReviewProgressReviewPandasViewSet,
    ReviewQuestionReviewNoCategoryModelViewSet,
    ReviewReviewGroupModelViewSet,
    ReviewTemplateModelViewSet,
)

# =================================================================================================================
#                                              CUSTOM FACTORY
# =================================================================================================================


@receiver(get_custom_factory, sender=ReviewTemplateModelViewSet)
def receive_factory_review_template(sender, *args, **kwargs):
    return ReviewTemplateFactory


@receiver(get_custom_factory, sender=ReviewQuestionReviewNoCategoryModelViewSet)
def receive_factory_review_question_no_category(sender, *args, **kwargs):
    return ReviewQuestionNoCategoryFactory


@receiver(get_custom_factory, sender=KPIModelViewSet)
def receive_factory_kpi(sender, *args, **kwargs):
    return DefaultPersonKPIFactory


# =================================================================================================================
#                                              UPDATE KWARGS
# =================================================================================================================


@receiver(custom_update_kwargs, sender=AbsenceTypeCountEmployeeModelViewSet)
def receive_kwargs_employee_absence(sender, *args, **kwargs):
    if kwargs.get("obj_factory"):
        abs_day = kwargs.get("obj_factory")
        abs_day.request.status = AbsenceRequest.Status.APPROVED
        abs_day.request.save()
        abs_day.save()
        return {"employee_id": abs_day.request.employee.id, "request_id": abs_day.request.id}
    else:
        return {}


@receiver(custom_update_kwargs, sender=ReviewTemplateModelViewSet)
@receiver(custom_update_kwargs, sender=ReviewReviewGroupModelViewSet)
@receiver(custom_update_kwargs, sender=ReviewModelViewSet)
def receive_kwargs_review(sender, *args, **kwargs):
    if (review := kwargs.get("obj_factory")) and (user := kwargs.get("user")):
        review.moderator = user.profile
        review.save()
    return {}


@receiver(custom_update_kwargs, sender=ReviewProgressReviewPandasViewSet)
@receiver(custom_update_kwargs, sender=ReviewAnswerReviewPandasViewSet)
@receiver(custom_update_kwargs, sender=ReviewAnswerReviewNoCategoryModelViewSet)
@receiver(custom_update_kwargs, sender=ReviewAnswerReviewQuestionCategoryModelViewSet)
def receive_kwargs_review_answer_question_category(sender, *args, **kwargs):
    if question_id := kwargs.get("question_id"):
        question = ReviewQuestion.objects.get(id=question_id)
        if question.category:
            return {"review_id": question.review.id, "category_id": question.category.id}
        else:
            return {"review_id": question.review.id}
    return {}


@receiver(custom_update_kwargs, sender=AbsenceTablePandasViewSet)
def receive_kwargs_employee_human_resource_absence(sender, *args, **kwargs):
    return {"request_id": None, "profile": None, "user": None, "obj_factory": None}


# =================================================================================================================
#                                              UPDATE SERIALIZER
# =================================================================================================================


@receiver(get_custom_serializer, sender=KPIModelViewSet)
def receive_serializer_kpi(sender, *args, **kwargs):
    return KPIModelSerializer


def app_pre_migration(sender, app_config, **kwargs):
    cur = connection.cursor()
    cur.execute("CREATE EXTENSION IF NOT EXISTS btree_gist;")



from datetime import datetime, timedelta
from decimal import Decimal
from typing import TypeVar

from celery import shared_task
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.files.base import ContentFile
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from slugify import slugify
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import Person
from wbcore.contrib.documents.models import Document, DocumentType
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.markdown.utils import custom_url_fetcher
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.models.fields import YearField
from wbcore.models.orderable import OrderableModel
from wbcore.signals import pre_merge
from wbcore.utils.models import CloneMixin, ComplexToStringMixin
from wbcore.workers import Queue
from weasyprint import HTML

from wbhuman_resources.models.employee import get_main_company
from wbhuman_resources.models.kpi import Evaluation

SelfReview = TypeVar("SelfReview", bound="Review")


def can_trigger_review(instance, user):
    return user.profile == instance.moderator and not instance.is_template and instance.reviewee and instance.reviewer


def can_validate_review(instance, user):
    return can_trigger_review(instance, user) and instance.signed_reviewee and instance.signed_reviewer


class ReviewGroup(WBModel):
    class Meta:
        verbose_name = _("Review Group")
        verbose_name_plural = _("Review Groups")

    name = models.CharField(max_length=255, verbose_name=_("Name"))
    employees = models.ManyToManyField("directory.Person", related_name="reviewgroups", blank=True)

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewgroup"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewgrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class Review(CloneMixin, ComplexToStringMixin, WBModel):
    class Status(models.TextChoices):
        PREPARATION_OF_REVIEW = "PREPARATION_OF_REVIEW", _("Stage 1: Preparation of review")
        FILL_IN_REVIEW = "FILL_IN_REVIEW", _("Stage 2: Fill in review")
        REVIEW = "REVIEW", _("Stage 3: Review")
        EVALUATION = "EVALUATION", _("Stage 4: Evalutation")
        VALIDATION = "VALIDATION", _("Stage 5: Validation")

        @classmethod
        def get_color_map(cls):
            colors = [
                WBColor.BLUE_LIGHT.value,
                "rgb(230,230,250)",
                WBColor.YELLOW_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
                WBColor.GREEN_DARK.value,
            ]
            return [choice for choice in zip(cls, colors, strict=False)]

    class Type(models.TextChoices):
        ANNUAL = "ANNUAL", _("Annual")
        INTERMEDIARY = "INTERMEDIARY", _("Intermediary")

    class Meta:
        verbose_name = _("Review")
        verbose_name_plural = _("Reviews")
        permissions = [
            ("administrate_review", "Can Administrate Reviews"),
        ]

        notification_types = [
            create_notification_type(
                code="wbhuman_resources.review.notify",
                title="Review Notification",
                help_text="Notifies you when a review has been submitted",
            )
        ]

    from_date = models.DateField(null=True, blank=True, verbose_name=_("From"))
    to_date = models.DateField(null=True, blank=True, verbose_name=_("To"))
    review_deadline = models.DateField(null=True, blank=True, verbose_name=_("Deadline"))
    review = models.DateTimeField(null=True, blank=True, verbose_name=_("Review Date"))
    auto_apply_deadline = models.BooleanField(default=True, verbose_name=_("Auto Apply Deadline"))
    year = YearField(null=True, blank=True, verbose_name=_("Year"))
    type = models.CharField(max_length=32, choices=Type.choices, default=Type.ANNUAL, verbose_name=_("Type"))
    status = FSMField(
        default=Status.PREPARATION_OF_REVIEW,
        choices=Status.choices,
        verbose_name=_("Status"),
        help_text=_("Indicates one of the four stages defined by the workflow"),
    )

    reviewee = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="reviewee_reviews",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Reviewee"),
    )
    reviewer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="reviewer_reviews",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Reviewer"),
    )

    moderator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="moderator_reviews",
        on_delete=models.deletion.SET_NULL,
    )

    review_group = models.ForeignKey(
        to="wbhuman_resources.ReviewGroup",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="review_related",
        verbose_name=_("Group"),
    )

    is_template = models.BooleanField(default=False)
    feedback_reviewee = models.TextField(default="", blank=True, verbose_name=_("Feedback Reviewee"))
    feedback_reviewer = models.TextField(default="", blank=True, verbose_name=_("Feedback Reviewer"))

    signed_reviewee = models.DateTimeField(null=True, blank=True, verbose_name=_("Date of reviewee's signature"))
    signed_reviewer = models.DateTimeField(null=True, blank=True, verbose_name=_("Date of reviewer's signature"))

    completely_filled_reviewee = models.DateTimeField(
        null=True, blank=True, verbose_name=_("Completely Filled Out Reviewee")
    )
    completely_filled_reviewer = models.DateTimeField(
        null=True, blank=True, verbose_name=_("Completely Filled Out Reviewer")
    )

    changed = models.DateTimeField(auto_now=True, null=True, blank=True)

    def compute_str(self) -> str:
        _str = ""
        if self.is_template:
            _str += (
                _("{group} [Template]").format(group=str(self.review_group)) if self.review_group else _("[Template]")
            )
        else:
            _str += (
                _("{reviewee}'s Review").format(reviewee=self.reviewee.computed_str) if self.reviewee else _("Review")
            )
        if self.from_date and self.to_date:
            _str += f" - ({self.from_date} - {self.to_date})"
        elif self.year:
            _str += f" - ({self.year})"
        return _str

    def __str__(self):
        return self.computed_str

    def save(self, *args, **kwargs):
        if self.reviewee and (employee := getattr(self.reviewee, "human_resources", None)) and not self.reviewer:
            self.reviewer = next(employee.get_managers(only_direct_manager=True), None)
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @property
    def is_cloneable(self) -> bool:
        """
        Property used by the CloneMixin to disable the instance cloning functionality from the view
        """
        return self.is_template

    def _clone(self, **kwargs) -> SelfReview:
        """
        Create new row in database with the same data as original instance have.
        We need to clone also the related objects and set correct foreign key value
        """
        object_copy = Review.objects.get(id=self.id)
        object_copy.id = None
        object_copy.save()
        for related_object in self.questions.all():
            related_object.review = object_copy
            related_object.id = None
            related_object.save()
        return object_copy

    @classmethod
    def get_administrators(cls) -> QuerySet[User]:
        return User.objects.filter(
            Q(groups__permissions__codename="administrate_review")
            | Q(user_permissions__codename="administrate_review")
        ).distinct()

    @transition(
        field=status,
        source=Status.PREPARATION_OF_REVIEW,
        target=Status.FILL_IN_REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label=_("Submit"),
                action_label=_("Stage 2: Fill in review"),
                description_fields=_(
                    "<p>Status: <b>Stage 1: Preparation of review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to submit this review to <b>Stage 2: Fill in review?</b></p>"
                ),
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        submit_review.delay(self.id)

    @transition(
        field=status,
        source=Status.FILL_IN_REVIEW,
        target=Status.REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="finalize",
                label=_("Finalize"),
                action_label=_("Stage 3: Review"),
                description_fields=_(
                    "<p>Status: <b>Stage 2: Fill in review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 3: Review?</b></p>"
                ),
            )
        },
    )
    def finalize(self, by=None, description=None, **kwargs):
        if not self.review_deadline:
            self.review_deadline = datetime.now().date()
        finalize_review.delay(self.id)
        if not by:
            self.send_review_notification(
                title=gettext("Stage 3: Review - {self}").format(self=str(self)),
                message=gettext("{self} has moved to stage 3: Review. You can now organize the evaluation").format(
                    self=str(self)
                ),
                recipient=self.moderator.user_account,
                message_alert_deadline=False,
            )

    @transition(
        field=status,
        source=Status.REVIEW,
        target=Status.FILL_IN_REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.EDIT.icon,
                key="undo",
                label=_("Undo"),
                action_label=_("Stage 2: Fill in review"),
                description_fields=_(
                    "<p>Status: <b>Stage 3: Review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 2: Fill in review?</b></p>"
                ),
            )
        },
    )
    def undo(self, by=None, description=None, **kwargs):
        if self.review_deadline <= datetime.now().date():
            self.review_deadline = datetime.now().date() + timedelta(days=1)

    @transition(
        field=status,
        source=Status.REVIEW,
        target=Status.EVALUATION,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="evaluate",
                label=_("Evaluate"),
                action_label=_("Stage 4: Evaluation"),
                description_fields=_(
                    "<p>Status: <b>Stage 3: Review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 4: Evaluation?</b></p>"
                ),
            )
        },
    )
    def evaluate(self, by=None, description=None, **kwargs):
        if not self.review:
            self.review = timezone.now()

    @transition(
        field=status,
        source=Status.EVALUATION,
        target=Status.VALIDATION,
        permission=can_validate_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="validation",
                label=_("Validate"),
                action_label=_("Stage 4: Evaluation"),
                description_fields=_(
                    "<p>Status: <b>Stage 4: Evaluation</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Do you want to send this review to <b>Stage 5: Validation?</b></p>"
                ),
            )
        },
    )
    def validation(self, by=None, description=None, **kwargs):
        users = User.objects.filter(
            Q(groups__permissions__codename="administrate_review")
            | Q(user_permissions__codename="administrate_review")
        ).distinct()
        for user in users:
            send_review_report_via_mail.delay(user.id, self.id)

    def send_review_notification(self, title, message, recipient, message_alert_deadline=True):
        if message_alert_deadline:
            message += gettext(
                "<p>Please pay particular attention to the deadline <b>{deadline}</b>, the review will automatically move to step 3 of the workflow on that date. At this stage the data will be frozen, you will not be able to modify it.</p>"
            ).format(deadline=self.review_deadline)

        send_notification(
            code="wbhuman_resources.review.notify",
            title=title,
            body=message,
            user=recipient,
            reverse_name="wbhuman_resources:review-detail",
            reverse_args=[self.id],
        )

    def generate_pdf(self):
        html = get_template("review/review_report.html")

        text_answers = ReviewAnswer.objects.filter(
            Q(question__review=self) & Q(question__answer_type=ReviewQuestion.ANSWERTYPE.TEXT)
        ).order_by("question__category__order", "question__order")
        category_ids = text_answers.values_list("question__category", flat=True).distinct()

        table = {}
        for category_id in category_ids:
            if category_id:
                category = ReviewQuestionCategory.objects.get(id=category_id)
                table[category.id] = {"name": category.name, "questions": {}}
            else:
                category_id = ""
                category = None
                table[""] = {"name": "", "questions": {}}
            question_ids = (
                text_answers.filter(question__category=category)
                .order_by("question__order")
                .values_list("question", flat=True)
                .distinct()
            )
            for question_id in question_ids:
                question = ReviewQuestion.objects.get(id=question_id)
                table[category_id]["questions"][question_id] = {"name": question.question, "answers": {}}
                for answer in text_answers.filter(
                    Q(question__category=category) & Q(question__id=question_id)
                ).order_by("question__order", "answered_by"):
                    table[category_id]["questions"][question_id]["answers"][answer.answered_by] = answer.answer_text

        rating_answers = (
            ReviewAnswer.objects.filter(Q(question__review=self) & Q(question__answer_type="RATING"))
            .annotate(
                is_reviewee=Case(
                    When(answered_by=self.reviewee, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
                is_reviewer=Case(
                    When(answered_by=self.reviewer, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
            )
            .order_by("question__category__order")
        )
        category_ids = rating_answers.values_list("question__category", flat=True).distinct()

        rating_table = {}
        for category_id in category_ids:
            if category_id:
                category = ReviewQuestionCategory.objects.get(id=category_id)
                rating_table[category.id] = {"name": category.name, "questions": {}}
            else:
                category_id = ""
                category = None
                rating_table[""] = {"name": "", "questions": {}}
            question_ids = (
                rating_answers.filter(question__category=category)
                .order_by("question__order")
                .values_list("question", flat=True)
                .distinct()
            )
            for question_id in question_ids:
                question = ReviewQuestion.objects.get(id=question_id)
                rating_table[category_id]["questions"][question_id] = {
                    "name": question.question,
                    "answers": {},
                    "answers_text": {},
                }
                for answer in rating_answers.filter(
                    Q(question__category=category) & Q(question__id=question_id)
                ).order_by("question__order", "answered_by"):
                    if answer.is_reviewee:
                        rating_table[category_id]["questions"][question_id]["answers"]["reviewee"] = (
                            answer.answer_number if answer.answer_number else "-"
                        )
                        rating_table[category_id]["questions"][question_id]["answers_text"]["reviewee"] = (
                            answer.answer_text if answer.answer_text else None
                        )
                    if answer.is_reviewer:
                        rating_table[category_id]["questions"][question_id]["answers"]["reviewer"] = (
                            answer.answer_number if answer.answer_number else "-"
                        )
                        rating_table[category_id]["questions"][question_id]["answers_text"]["reviewer"] = (
                            answer.answer_text if answer.answer_text else None
                        )

        total_reviewee = rating_answers.filter(is_reviewee=True).aggregate(total=Sum("question__weight"))["total"]
        total_reviewer = rating_answers.filter(is_reviewer=True).aggregate(total=Sum("question__weight"))["total"]
        global_rating_reviewee = (
            rating_answers.filter(is_reviewee=True)
            .annotate(global_rating=F("question__weight") * F("answer_number"))
            .aggregate(total=Sum("global_rating"))["total"]
        )
        global_rating_reviewer = (
            rating_answers.filter(is_reviewer=True)
            .annotate(global_rating=F("question__weight") * F("answer_number"))
            .aggregate(total=Sum("global_rating"))["total"]
        )
        global_rating_reviewee = (
            0 if total_reviewee is None or global_rating_reviewee is None else global_rating_reviewee / total_reviewee
        )
        global_rating_reviewer = (
            0 if total_reviewer is None or global_rating_reviewer is None else global_rating_reviewer / total_reviewer
        )

        html_content = html.render(
            {
                "base_url": Site.objects.get_current().domain,
                "review": self,
                "table": table,
                "rating_table": rating_table,
                "global_rating_reviewee": round(global_rating_reviewee, 2),
                "global_rating_reviewer": round(global_rating_reviewer, 2),
                "number_total_rating": ReviewQuestion.objects.filter(Q(review=self) & Q(answer_type="RATING")).count(),
            }
        )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    @classmethod
    def get_subquery_review_related_to(cls, requester):
        answers = ReviewAnswer.objects.filter(question__review=OuterRef("pk"), answered_by=requester).order_by()
        return Coalesce(Subquery(answers.values("question__review")[:1]), None)

    @classmethod
    def subquery_global_rating(cls, requester):
        answers = ReviewAnswer.objects.filter(
            question__review=OuterRef("pk"),
            answered_by=requester,
            question__answer_type=ReviewQuestion.ANSWERTYPE.RATING,
            answer_number__isnull=False,
        ).order_by()

        subquery = (
            answers.annotate(
                rating_weighted=F("question__weight") * F("answer_number"),
            )
            .values("question__review")
            .annotate(
                total_rating=Sum("rating_weighted"),
                total_weight=Sum("question__weight"),
            )
            .annotate(global_rating=Coalesce(F("total_rating") / F("total_weight"), Decimal(0)))
            .values("global_rating")[:1]
        )

        return Coalesce(Subquery(subquery), None)

    def get_question_categories(self) -> models.QuerySet:
        return ReviewQuestionCategory.objects.filter(questions_related__review=self).order_by("order").distinct()

    def get_answer_categories_for_user(self, user: User) -> models.QuerySet:
        if self.status in [self.Status.FILL_IN_REVIEW, self.Status.REVIEW]:
            return (
                ReviewQuestionCategory.objects.filter(
                    questions_related__answers__answered_by=user.profile, questions_related__review=self
                )
                .order_by("order")
                .distinct()
            )
        else:  # EVALUATION
            return (
                ReviewQuestionCategory.objects.filter(
                    questions_related__answer_type=ReviewQuestion.ANSWERTYPE.TEXT, questions_related__review=self
                )
                .order_by("order")
                .distinct()
            )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:review"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class ReviewQuestionCategory(OrderableModel, WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"), unique=True)
    weight = models.DecimalField(default=1.0, max_digits=16, decimal_places=1, verbose_name=_("Weight"))

    class Meta(OrderableModel.Meta):
        verbose_name = _("Review Question Category")
        verbose_name_plural = _("Review Question Categories")

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewquestioncategory"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewquestioncategoryrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class ReviewQuestion(ComplexToStringMixin, OrderableModel, WBModel):
    order_with_respect_to = ("review", "category")

    class ANSWERTYPE(models.TextChoices):
        TEXT = "TEXT", _("Text")
        RATING = "RATING", _("Rating")

    review = models.ForeignKey(
        to="wbhuman_resources.Review",
        on_delete=models.CASCADE,
        related_name="questions",
        verbose_name=_("Review"),
    )
    category = models.ForeignKey(
        to="wbhuman_resources.ReviewQuestionCategory",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="questions_related",
        verbose_name=_("Category"),
    )
    question = models.TextField(default="", blank=True, verbose_name=_("Question"))
    mandatory = models.BooleanField(default=True, verbose_name=_("Mandatory"))
    answer_type = models.CharField(
        max_length=32,
        default=ANSWERTYPE.TEXT,
        choices=ANSWERTYPE.choices,
        verbose_name=_("Type"),
    )
    for_reviewee = models.BooleanField(default=True, verbose_name=_("For Reviewee"))
    for_reviewer = models.BooleanField(default=True, verbose_name=_("For Reviewer"))
    for_department_peers = models.BooleanField(default=False, verbose_name=_("For Department Peers"))
    for_company_peers = models.BooleanField(default=False, verbose_name=_("For Company Peers"))
    weight = models.DecimalField(default=1.0, max_digits=16, decimal_places=1, verbose_name=_("Weight"))
    evaluation = models.ForeignKey(
        to="wbhuman_resources.Evaluation",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="evaluation_questions",
        verbose_name=_("Evaluation"),
    )

    class Meta:
        verbose_name = _("Review Question")
        verbose_name_plural = _("Review Questions")
        ordering = ("review", "category", "order")

    def compute_str(self) -> str:
        return f"{self.question}"

    def __str__(self):
        return f"({self.id}) - {self.computed_str}"

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewquestion"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewquestionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class ReviewAnswer(models.Model):
    class Meta:
        verbose_name = _("Review Answer")
        verbose_name_plural = _("Review Answers")

    question = models.ForeignKey(
        to="wbhuman_resources.ReviewQuestion",
        on_delete=models.CASCADE,
        related_name="answers",
        verbose_name=_("Question"),
    )

    answered_by = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="related_answers",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Answered By"),
    )

    answered_anonymized = models.CharField(
        null=True, blank=True, max_length=255, verbose_name=_("Answered Anonymized")
    )
    answer_number = models.PositiveIntegerField(
        null=True, blank=True, verbose_name=_("Rating"), validators=[MinValueValidator(1), MaxValueValidator(4)]
    )
    answer_text = models.TextField(null=True, blank=True, verbose_name=_("Comment"))

    def __str__(self) -> str:
        return f"{self.question} - {self.answered_by}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewanswer"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "id"


@receiver(pre_save, sender=Review)
def pre_save_review(sender, instance, **kwargs):
    if (
        not instance.is_template
        and instance.auto_apply_deadline
        and instance.review_deadline
        and instance.status == Review.Status.FILL_IN_REVIEW
    ):
        if instance.review_deadline < datetime.now().date():
            instance.finalize()

    if (
        not instance.is_template
        and instance.signed_reviewee
        and instance.signed_reviewer
        and instance.status == Review.Status.EVALUATION
    ):
        instance.validation()


@shared_task(queue=Queue.DEFAULT.value)
def finalize_review(review_id):
    review = Review.objects.get(id=review_id)
    ReviewAnswer.objects.filter(
        Q(question__review=review)
        & Q(answered_by=review.reviewee)
        & Q(question__answer_type=ReviewQuestion.ANSWERTYPE.RATING)
        & Q(answer_number=None)
    ).update(answer_number=1)


@shared_task(queue=Queue.DEFAULT.value)
def submit_reviews_from_group(group_id, user_id):
    user = User.objects.get(id=user_id)

    reviews = Review.objects.filter(
        Q(review_group__id=group_id)
        & Q(status=Review.Status.PREPARATION_OF_REVIEW)
        & Q(moderator=user.profile)
        & Q(is_template=False)
    )
    for review in reviews:
        review.submit()
        review.save()


@shared_task(queue=Queue.DEFAULT.value)
def create_review_from_template(
    template_id, from_date, to_date, review_deadline, auto_apply_deadline, employees, include_kpi
):
    template = Review.objects.get(id=template_id)

    from_date = datetime.strptime(from_date, "%Y-%m-%d").date() if from_date and from_date != "null" else None
    to_date = datetime.strptime(to_date, "%Y-%m-%d").date() if to_date and to_date != "null" else None
    review_deadline = (
        datetime.strptime(review_deadline, "%Y-%m-%d").date()
        if review_deadline and review_deadline != "null"
        else None
    )

    include_kpi = str(include_kpi).lower() in ("yes", "true", "t", "1") if include_kpi else False

    _auto_apply_deadline = (
        str(auto_apply_deadline).lower() in ("yes", "true", "t", "1")
        if auto_apply_deadline
        else template.auto_apply_deadline
    )
    list_employees = Person.objects.filter(id__in=employees.split(",")) if employees else []

    for employee in list_employees:
        review = Review.objects.create(
            from_date=from_date,
            to_date=to_date,
            review_deadline=review_deadline,
            reviewee=employee,
            moderator=template.moderator,
            auto_apply_deadline=_auto_apply_deadline,
            status=template.status,
            review_group=template.review_group,
            year=template.year,
            type=template.type,
        )

        for _question in ReviewQuestion.objects.filter(review__id=template.id).order_by("order"):
            kwargs = {
                "review": review,
                "category": _question.category,
                "question": _question.question,
                "answer_type": _question.answer_type,
                "mandatory": _question.mandatory,
                "for_reviewee": _question.for_reviewee,
                "for_reviewer": _question.for_reviewer,
                "for_department_peers": _question.for_department_peers,
                "for_company_peers": _question.for_company_peers,
                "order": _question.order,
                "weight": _question.weight,
            }
            ReviewQuestion.objects.create(**kwargs)

        if include_kpi:
            qs_evaluations = Evaluation.objects.filter(is_active=True)
            evaluations = (
                qs_evaluations.filter(
                    Q(person=employee) | (Q(person__isnull=True) & Q(kpi__evaluated_persons=employee))
                )
                .annotate(
                    goal=F("kpi__goal"),
                    kpi_name=F("kpi__name"),
                    interval=F("kpi__evaluated_intervals"),
                    period=F("kpi__period"),
                    person_name=F("person__computed_str"),
                )
                .order_by("kpi", "-evaluation_date")
                .distinct("kpi")
            )

            category, _ = ReviewQuestionCategory.objects.get_or_create(name="KPI Evaluations")
            for evaluation in evaluations:
                percentage = round((evaluation.evaluated_score / evaluation.goal) * 100, 2)
                person_name = evaluation.person_name if evaluation.person_name else "Group"
                _question = gettext(
                    """The KPI '{kpi_name}' is evaluated {interval} over the period from {lower_period} to {upper_period}.
{person_name}'s evaluation as of {evaluation_date} was {evaluated_score} for a goal of {goal}. That is an accomplished percentage of {percentage}%.
What do you think of this result?
                """
                ).format(
                    kpi_name=evaluation.kpi_name,
                    interval=evaluation.interval,
                    lower_period=evaluation.period.lower,
                    upper_period=evaluation.period.upper,
                    person_name=person_name,
                    evaluation_date=evaluation.evaluation_date,
                    evaluated_score=evaluation.evaluated_score,
                    goal=evaluation.goal,
                    percentage=percentage,
                )

                kwargs = {
                    "review": review,
                    "category": category,
                    "question": _question,
                    "answer_type": ReviewQuestion.ANSWERTYPE.RATING,
                    "evaluation": evaluation,
                }
                ReviewQuestion.objects.create(**kwargs)


@shared_task(queue=Queue.DEFAULT.value)
def submit_review(review_id):
    review = Review.objects.get(id=review_id)
    questions = ReviewQuestion.objects.filter(review=review)
    dict_questions = {}

    questions_for_company = questions.filter(for_company_peers=True)
    if questions_for_company.exists() and (main_company_id := get_main_company()):
        if _company := review.reviewee.employers.filter(id=main_company_id).first():
            for _person in (
                Person.objects.filter_only_internal()
                .filter(employers=_company)
                .exclude(Q(id=review.reviewee.id) | Q(id=review.reviewer.id))
            ):
                dict_questions[_person] = questions_for_company

    questions_for_department = questions.filter(for_department_peers=True)
    if questions_for_department.exists() and hasattr(review.reviewee, "human_resources"):
        if review.reviewee.human_resources.position:
            employees_peers = []
            for _position in review.reviewee.human_resources.position.get_ancestors(include_self=True):
                employees_peers += _position.get_employees().exclude(
                    Q(id=review.reviewee.id) & Q(id=review.reviewer.id)
                )
            person_peers = [_employee.profile for _employee in list(set(employees_peers))]
            for _peer in list(set(person_peers)):
                if _peer in dict_questions:
                    dict_questions[_peer] = questions_for_department.union(questions_for_company)
                else:
                    dict_questions[_peer] = questions_for_department

    dict_questions[review.reviewee] = questions.filter(for_reviewee=True)
    dict_questions[review.reviewer] = questions.filter(for_reviewer=True)

    for person, values in dict_questions.items():
        for _question in values:
            kwargs = {"question": _question, "answered_by": person, "answered_anonymized": hash(person)}
            if _evaluation := _question.evaluation:
                rating = _evaluation.get_rating()
                kwargs.update({"answer_number": rating})
            ReviewAnswer.objects.create(**kwargs)
        if hasattr(person, "user_account"):
            if person == review.reviewee:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>your review is ready. You can fill it in until {deadline}.</p>"
                ).format(first_name=person.first_name, last_name=person.last_name, deadline=review.review_deadline)
            elif person == review.reviewer:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>you are the reviewer of the review: {review}. You can fill in the questions related to you until <b>{deadline}.</b></p>"
                ).format(
                    first_name=person.first_name,
                    last_name=person.last_name,
                    review=str(review),
                    deadline=review.review_deadline,
                )
            else:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>you are a peer of the review: {review}. You can fill in the questions related to you until <b>{deadline}.</b></p>"
                ).format(
                    first_name=person.first_name,
                    last_name=person.last_name,
                    review=str(review),
                    deadline=review.review_deadline,
                )

            review.send_review_notification(
                title=gettext("Stage 2: Fill in review - {review}").format(review=str(review)),
                message=msg,
                recipient=person.user_account,
            )


@shared_task(queue=Queue.DEFAULT.value)
def send_review_report_via_mail(user_id, review_id):
    user = User.objects.get(id=user_id)
    review = Review.objects.get(id=review_id)
    pdf_content = review.generate_pdf()
    filename = f"{slugify(str(review))}.pdf"

    content_file = ContentFile(pdf_content, name=filename)
    document_type, _ = DocumentType.objects.get_or_create(name="mailing")

    document, _ = Document.objects.update_or_create(
        document_type=document_type,
        system_created=True,
        system_key=f"review-{review.id}-{filename}",
        defaults={
            "file": content_file,
            "name": filename,
            "permission_type": Document.PermissionType.PRIVATE,
            "creator": user,
        },
    )
    document.link(review)

    document.send_email(to_emails=user.email, as_link=True, subject=gettext("Review PDF - ") + review.computed_str)


@receiver(pre_merge, sender="directory.Person")
def pre_merge_person_review(sender: models.Model, merged_object, main_object, **kwargs):
    ReviewAnswer.objects.filter(answered_by=merged_object).update(answered_by=main_object)
    Review.objects.filter(reviewee=merged_object).update(reviewee=main_object)
    Review.objects.filter(reviewer=merged_object).update(reviewer=main_object)
    Review.objects.filter(moderator=merged_object).update(moderator=main_object)

    for group in ReviewGroup.objects.filter(employees=merged_object):
        group.employees.remove(merged_object)
        if main_object not in group.employees.all():
            group.employees.add(main_object)
